///////////////////////////////////////////////////////////////////////////////
//    
// ()  ...,  ""
///////////////////////////////////////////////////////////////////////////////
//   ""
///////////////////////////////////////////////////////////////////////////////
#include "KAMAMeas.h"
#include "GreenwichFrame.h"
#include "PhasePoint.h"
#include "Consts.h"
#include "TimeFunctions.h"

KAMAMeasurement::KAMAMeasurement() : IMeasurement()
{
	Type = mtRTKama;
	m_isBad = false;
}

///////////////////////////////////////////////////////////////////////////////
//   ""
///////////////////////////////////////////////////////////////////////////////
//       ,
//       trj.
//    :
// t	-	,     
//           (    ,
//          )
// fr	-	 ,       
//        ,      (.   f  m)
// v	-	   .     0,
//          
// m	-	   Nx6,  N -   
//      (..    N   6 ).   
//        Rx Ry Rz Vx  Vy Vz,    -   .
//          0,     
// del	-	   .     0,
//           
//-----------------------------------------------------------------------------
//       (1 )
#define EPS_D (1.e-8)
bool KAMAMeasurement::Simulate(int matrixSize, ITraject &trj, TimeJD &t, const IFrame &fr, Vect6 *f,
							   DVector *v, DMatrix *m, DVector *del) const
{
	//        
	//  ,     ( J2000)

	PhasePoint3D gr, rec, snd;
	GreenwichFrame gf;
		
	gr.Init(gf, PRec, MTime);
	rec.SetFrame(fr);
	rec.Assign(gr);

	//         
	// c   EPS_D

	double tau2, tau, d1, d2;
	Vect6 fo;

	if (f == 0)
		f = &fo;

	tau2 = 0;
	for (;;)
	{
		t = MTime;
		t += tau2 / 86.4;
		if (!trj.Predict(t, fr, *f))
			return false;
		d2 = sqrt((rec.Coords.x - f->x) * (rec.Coords.x - f->x) +
			(rec.Coords.y - f->y) * (rec.Coords.y - f->y) +
			(rec.Coords.z - f->z) * (rec.Coords.z - f->z));
		tau = d2 / Clight;
		if (fabs(tau - tau2) * Clight < EPS_D)
			break;
		tau2 = tau;
	}

	TimeJD t1;
	double tau1;

	//       
	//     

	tau1 = 0;
	for (;;)
	{
		t = MTime;
		t += tau2 / 86.4;
		t += tau1 / 86.4;
		gr.Init(gf, PSend, t);
		snd.SetFrame(fr);
		snd.Assign(gr);
		d1 = sqrt((snd.Coords.x - f->x) * (snd.Coords.x - f->x) +
			(snd.Coords.y - f->y) * (snd.Coords.y - f->y) +
			(snd.Coords.z - f->z) * (snd.Coords.z - f->z));
		tau = d1 / Clight;
		if (fabs(tau - tau1) * Clight < EPS_D)
			break;
		tau1 = tau;
	}

	if (v != NULL)
	{
		v->Init(1U);
		(*v)(0) = 0.5 * (d1 + d2);
	}

	if (del != NULL)
	{
		del->Init(1U);
		(*del)(0) = D - 0.5 * (d1 + d2);
	}

	if (m != NULL)
	{
		m->Init(1, 6);
		int i;
		for (i = 0; i < 3; i++)
		{
			(*m)(0, i) = 0.5 * (((*f)[i] - rec.Coords[i]) / d2 + ((*f)[i] - snd.Coords[i]) / d1);
			(*m)(0, i + 3) = 0.0;
		}
	}

	return true;
};
//-----------------------------------------------------------------------------
//      new()
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void KAMAMeasurement::Diff(DVector &d, const DVector *m1, const DVector *m2) const
{
	DVector s1(1), s2(1);

	s1(0) = m1 == NULL ? D : (*m1)(0);
	s2(0) = m2 == NULL ? D : (*m2)(0);

	d.Init(1U);
	d(0) = s1(0) - s2(0);
}
//-----------------------------------------------------------------------------
//      new()
//-----------------------------------------------------------------------------
IMeasurement* KAMAMeasurement::Clone() const
{
	KAMAMeasurement *m = new KAMAMeasurement;
	*m = *this;
	return m;
}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
//       ,
//       trj.
//    :
// t	-	,     
//           (    ,
//          )
// fr	-	 ,       
//        ,      (.   f  m)
// v	-	   .     0,
//          
// m	-	   Nx6,  N -   
//      (..    N   6 ).   
//        Rx Ry Rz Vx  Vy Vz,    -   .
//          0,     
// del	-	   .     0,
//           
bool KAMADtMeasurement::Simulate(int matrixSize, ITraject &trj, TimeJD &t, const IFrame &fr, Vect6 *f,
								 DVector *v, DMatrix *m, DVector *del) const
{
	//        
	//  ,     ( J2000)

	PhasePoint6D gr, rec, snd;
	GreenwichFrame gf;
	Vect3 v0(0.0, 0.0, 0.0);

	gr.Init(gf, PRec, v0, MTime);
	rec.SetFrame(fr);
	rec.Assign(gr);

	//         
	// c   EPS_D

	double tau2, tau, d1, d2;
	Vect6 fo;
	Vect3 r1, r2, v1, v2;

	if (f == 0)
		f = &fo;

	tau2 = 0;
	for (;;)
	{
		t = MTime;
		t -= tau2 / 86.4;
		if (!trj.Predict(t, fr, *f))
			return false;
		r2.x = f->x - rec.CoordsVel.x;
		r2.y = f->y - rec.CoordsVel.y;
		r2.z = f->z - rec.CoordsVel.z;
		d2 = r2.Len();
		tau = d2 / Clight;
		if (fabs(tau - tau2) * Clight < EPS_D)
			break;
		tau2 = tau;
	}
	v2.x = f->vx - rec.CoordsVel.vx;
	v2.y = f->vy - rec.CoordsVel.vy;
	v2.z = f->vz - rec.CoordsVel.vz;

	TimeJD t1;
	double tau1;

	//       
	//     

	tau1 = 0;
	for (;;)
	{
		t = MTime;
		t -= tau2 / 86.4;
		t -= tau1 / 86.4;
		gr.Init(gf, PSend, v0, t);
		snd.SetFrame(fr);
		snd.Assign(gr);
		r1.x = f->x - snd.CoordsVel.x;
		r1.y = f->y - snd.CoordsVel.y;
		r1.z = f->z - snd.CoordsVel.z;
		d1 = r1.Len();
		tau = d1 / Clight;
		if (fabs(tau - tau1) * Clight < EPS_D)
			break;
		tau1 = tau;
	}
	v1.x = f->vx - snd.CoordsVel.vx;
	v1.y = f->vy - snd.CoordsVel.vy;
	v1.z = f->vz - snd.CoordsVel.vz;

	double d1t, d2t;

	d1t = (r1 * v1) / d1;
	d2t = (r2 * v2) / d2;

	if (v != NULL)
	{
		v->Init(1U);
		(*v)(0) = 0.5 * (d1t + d2t);
	}

	if (del != NULL)
	{
		del->Init(1U);
		(*del)(0) = Dt - 0.5 * (d1t + d2t);
	}

	if (m != NULL)
	{
		m->Init(1, 6);
		int i;
		for (i = 0; i < 3; i++)
		{
			(*m)(0, i) = 0.5 * ((v1[i] - d1t / d1 * r1[i]) / d1 + (v2[i] - d2t / d2 * r2[i]) / d2);
			(*m)(0, i + 3) = 0.5 * (r1[i] / d1 + r2[i] / d2);
		}
	}

	return true;
}
//-----------------------------------------------------------------------------
//      new()
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void KAMADtMeasurement::Diff(DVector &d, const DVector *m1, const DVector *m2) const
{
	DVector s1(1), s2(1);

	s1(0) = m1 == NULL ? Dt : (*m1)(0);
	s2(0) = m2 == NULL ? Dt : (*m2)(0);

	d.Init(1U);
	d(0) = s1(0) - s2(0);
}
//-----------------------------------------------------------------------------
//      new()
//-----------------------------------------------------------------------------
IMeasurement* KAMADtMeasurement::Clone() const
{
	KAMADtMeasurement *m = new KAMADtMeasurement;
	*m = *this;
	return m;
}
///////////////////////////////////////////////////////////////////////////////

