///////////////////////////////////////////////////////////////////////////////
//    
// ()  ...,  ""
///////////////////////////////////////////////////////////////////////////////
//     
////////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_SFRAMEPARS_H
#define BNOTOOL_SFRAMEPARS_H
////////////////////////////////////////////////////////////////////////////////
#include "IFrames.h"
#include "Defines.h"
#include "GeoCoords.h"
////////////////////////////////////////////////////////////////////////////////
//     ,     
////////////////////////////////////////////////////////////////////////////////
class SFrameParameters : public IFrameParameters
  {
  public:
      //    
    virtual TCoordSysId GetFrameType() const
      { return FrameType;};
      //    
    virtual bool IsEqual(const IFrameParameters &c) const
      { return FrameType == c.GetFrameType(); };
  
  protected:
    TCoordSysId FrameType;
  };
////////////////////////////////////////////////////////////////////////////////
//   J2000 ()
////////////////////////////////////////////////////////////////////////////////
class J2000FrameParameters : public SFrameParameters
  {
  public:
    J2000FrameParameters()
      { FrameType = csIdJ2000;};
  };
////////////////////////////////////////////////////////////////////////////////
//   J2000 ()
////////////////////////////////////////////////////////////////////////////////
class J2000SFrameParameters : public SFrameParameters
  {
  public:
    J2000SFrameParameters()
      { FrameType = csIdJ2000S;};
  };
////////////////////////////////////////////////////////////////////////////////
//   B1950
////////////////////////////////////////////////////////////////////////////////
class B1950FrameParameters : public SFrameParameters
  {
  public:
    B1950FrameParameters()
      { FrameType = csIdB1950;};
  };
////////////////////////////////////////////////////////////////////////////////
//   B1900
////////////////////////////////////////////////////////////////////////////////
class B1900FrameParameters : public SFrameParameters
  {
  public:
    B1900FrameParameters()
      { FrameType = csIdB1900;};
  };
////////////////////////////////////////////////////////////////////////////////
//    
////////////////////////////////////////////////////////////////////////////////
class GreenwichFrameParameters : public SFrameParameters
  {
  public:
    GreenwichFrameParameters()
      { FrameType = csIdGCS;};
  };
////////////////////////////////////////////////////////////////////////////////
//        
////////////////////////////////////////////////////////////////////////////////
class TFrameParameters : public SFrameParameters
  {
  public:
      //   
    TFrameParameters();
      //      
    TFrameParameters(const TimeJD &t)
      { Init(t);};

      //    
    virtual bool IsEqual(const IFrameParameters &c) const;

      //     
    void Init(const TimeJD &t);

      //   
    const TimeJD &GetTime() const
      { TAssert(InitFlag); return TFix;};
      //  
    bool IsInit() const
      { return InitFlag;};

  protected:
    TimeJD TFix;
    bool InitFlag;
  };
////////////////////////////////////////////////////////////////////////////////
//   METE
////////////////////////////////////////////////////////////////////////////////
class METEFrameParameters : public TFrameParameters
  {
  public:
    METEFrameParameters() : TFrameParameters()
      { FrameType = csIdMETE;};
    METEFrameParameters(const TimeJD &t) : TFrameParameters(t)
      { FrameType = csIdMETE;};

  };
////////////////////////////////////////////////////////////////////////////////
//  ""   
////////////////////////////////////////////////////////////////////////////////
class FGreenwichFrameParameters : public TFrameParameters
  {
  public:
    FGreenwichFrameParameters() : TFrameParameters()
      { FrameType = csIdGFCS;};
    FGreenwichFrameParameters(const TimeJD &t) : TFrameParameters(t)
      { FrameType = csIdGFCS;};
  };
////////////////////////////////////////////////////////////////////////////////
//  ""   
////////////////////////////////////////////////////////////////////////////////
class ECIFrameParameters : public TFrameParameters
  {
  public:
    ECIFrameParameters() : TFrameParameters()
      { FrameType = csIdECI;};
    ECIFrameParameters(const TimeJD &t) : TFrameParameters(t)
      { FrameType = csIdECI;};
  };
////////////////////////////////////////////////////////////////////////////////
//    NEZ
////////////////////////////////////////////////////////////////////////////////
class NZEFrameParameters : public SFrameParameters
  {
  public:
      //   
    NZEFrameParameters()
      { Init(); InitFlag = false; };
      //        
    NZEFrameParameters(const Vect3 &p)
      { Init(p);};

      //    
    virtual bool IsEqual(const IFrameParameters &c) const;

      //        
    void Init(const Vect3 &p)
      { Init(); Point = p; InitFlag = true;};

      //   
    const Vect3 &GetPoint() const
      { TAssert(InitFlag); return Point;};
      //  
    bool IsInit() const
      { return InitFlag;};

  protected:
    void Init()
      { FrameType = csIdNZE;};

    Vect3 Point;
    bool InitFlag;
  };
////////////////////////////////////////////////////////////////////////////////
//    RNB
////////////////////////////////////////////////////////////////////////////////
class RNBFrameParameters : public SFrameParameters
  {
  public:
      //   
    RNBFrameParameters()
      { Init(); InitFlag = false; };
      //      
      //      J2000
    RNBFrameParameters(const Vect3 &p, const Vect3 &v)
      { Init(p, v);};
    RNBFrameParameters(const Vect6 &rv)
      { Init(rv);};

      //    
    virtual bool IsEqual(const IFrameParameters &c) const;

      //      
      //      J2000
    void Init(const Vect3 &p, const Vect3 &v);
    void Init(const Vect6 &rv);

      //   
    const Vect3 &GetPoint() const
      { TAssert(InitFlag); return Point;};
      //  
    bool IsInit() const
      { return InitFlag;};
    const TransMtx & GetMatrix() const
      { TAssert(InitFlag); return Matrix;};

  protected:
    void Init()
      { FrameType = csIdRNB;};

    Vect3 Point;      //     J2000
    bool InitFlag;
    TransMtx Matrix;  //    J2000  RNB
  };
////////////////////////////////////////////////////////////////////////////////
//    J2000
////////////////////////////////////////////////////////////////////////////////
class LocalJ2000FrameParameters : public SFrameParameters
  {
  public:
      //   
    LocalJ2000FrameParameters()
      { Init(); InitFlag = false; };
      //      
    LocalJ2000FrameParameters(const GeoCoordsPrec &p)
      { Init(p);};

      //    
    virtual bool IsEqual(const IFrameParameters &c) const;

      //      
    void Init(const GeoCoordsPrec &p)
      { GPoint = p; InitFlag = true;};

      //  
    bool IsInit() const
      { return InitFlag;};

      //     
    const GeoCoordsPrec &GetPoint() const
      { return GPoint;};

  protected:
    void Init()
      { FrameType = csIdLocalJ2000;};

    GeoCoordsPrec GPoint;   //    
    bool InitFlag;
  };
////////////////////////////////////////////////////////////////////////////////
//     
////////////////////////////////////////////////////////////////////////////////
class LocalRLFrameParameters : public SFrameParameters
  {
  public:
      //   
    LocalRLFrameParameters()
      { Init(); InitFlag = false; };
      //      
      //      J2000
    LocalRLFrameParameters(const GeoCoordsPrec &p, double az)
      { Init(p, az);};

      //    
    virtual bool IsEqual(const IFrameParameters &c) const;

      //      
    void Init(const GeoCoordsPrec &p, double az)
      { GPoint = p; Az = az; InitFlag = true;};

      //  
    bool IsInit() const
      { return InitFlag;};

      //     
    const GeoCoordsPrec &GetPoint() const
      { return GPoint;};
      //   
    double GetAz() const
      { return Az;};

  protected:
    void Init()
      { FrameType = csIdLocalRL;};

      //    
    GeoCoordsPrec GPoint;   
      //  []
    double Az;
    bool InitFlag;
  };
////////////////////////////////////////////////////////////////////////////////
#endif

