///////////////////////////////////////////////////////////////////////////////
//    
// ()  ...,  ""
///////////////////////////////////////////////////////////////////////////////
//     ( ).
//       .
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_EPHEMERIDE_H
#define BNOTOOL_EPHEMERIDE_H
///////////////////////////////////////////////////////////////////////////////
#include "TM.h"
#include "DVectors.h"
#include "IAstroBase.h"
///////////////////////////////////////////////////////////////////////////////
   //     ( ).
class Ephemeride : public IEphemeride
  {
  public:
      //     .
      //    .
    enum EnBodyID
      {
      idMercury = 0,
      idVenus = 1,
      idEarthMoon = 2,
      idMars = 3,
      idJupiter = 4,
      idSaturn = 5,
      idUranus = 6,
      idNeptune = 7,
      idPluto = 8,
      idMoon = 9,
      idSun = 10,
      idUnknown = -1
      };

      // .
    Ephemeride();
      //    
    Ephemeride(const char *fname, EnBodyID body)
      { XCoefs = NULL; Init(fname, body);};

      // .
    virtual ~Ephemeride();

      // 
    virtual void Init(const char *fname, EnBodyID body)
      { UnLoadData(); strcpy(FName, fname); BodyID = body;};

      //     
    virtual bool LoadData(const TimeJD &s, const TimeJD &e);

      //  
    virtual void UnLoadData();

      //   .
    virtual void GetPosition(const TimeJD &t, Vect3 *pos = NULL, 
                             Vect3 *vel = NULL, Vect3 *acc = NULL) const;
      //   .
    EnBodyID GetBodyID() const
      { return BodyID; }
    //      .
    const TimeJD& GetStartTime() const
      { return StartTime; };
	  //      .
    const TimeJD& GetEndTime() const
      { return EndTime; };
      //       .
    const TimeJD & GetFileStartTime() const
      { return FileStartTime; };
      //       .
    const TimeJD & GetFileEndTime() const
      { return FileEndTime; };

  protected:
      //  .
    EnBodyID BodyID;
      //    
    char FName[256];

  private:
      //   ,  32 .
    long D32Len;
      //      32- .
    long BodyOffset;
      //      .
    long CfLen;
      //  .
    TimeJD StartTime;
      //  .
    TimeJD EndTime;
      //     .
    TimeJD FileStartTime;
      //     .
    TimeJD FileEndTime;   
      //   .
    double *XCoefs, *YCoefs, *ZCoefs;

      //    ,  
      //  .
    bool ReadData(const TimeJD &start, const TimeJD &end);
      //   ,     
      //  .
    void CalculatePolynom(double t_cheb, double *coeff, double *coord,
                          double *coord_dt, double *coord_dt2) const;

    //   
      //   
    mutable TimeJD Tp, Tv, Ta;
    mutable Vect3 Cp, Cv, Ca;
      //   
    bool GetFromCash(const TimeJD &t, Vect3 *pos, Vect3 *vel, Vect3 *acc) const;
      //   
    void PutToCash(const TimeJD &t, Vect3 *pos, Vect3 *vel, Vect3 *acc) const;
      //  
    void ClearCash()
      { Tp = Tv = Ta = TimeJD(0, 0.0);};

      //  .
	  Ephemeride(const Ephemeride& src) ;
      //  .
	  Ephemeride& operator=(const Ephemeride& src) ;
  } ;
///////////////////////////////////////////////////////////////////////////////
//   
///////////////////////////////////////////////////////////////////////////////
class PlanetEphemeride : public Ephemeride
  {
  public:
      // .
    PlanetEphemeride() : Ephemeride()
      {};
      //    
    PlanetEphemeride(const char *fname, EnBodyID body) : Ephemeride(fname, body)
      {};

      //       
    virtual TCoordSysId GetCS() const
      { return csIdJ2000S;};
  };
///////////////////////////////////////////////////////////////////////////////
//   
///////////////////////////////////////////////////////////////////////////////
class MoonEphemeride : public Ephemeride
  {
  public: 
      // .
    MoonEphemeride() : Ephemeride()
      {BodyID = idMoon;};
      //    
    MoonEphemeride(const char *fname) : Ephemeride(fname, idMoon)
      {};

      //       
    virtual TCoordSysId GetCS() const
      { return csIdJ2000;};

      // 
    virtual void Init(const char *fname)
      { Ephemeride::Init(fname, idMoon);};
      // 
    virtual void Init(const char *fname, EnBodyID body)
      { Ephemeride::Init(fname, idMoon);};
  };
///////////////////////////////////////////////////////////////////////////////
//   
///////////////////////////////////////////////////////////////////////////////
class SunEphemeride : public IEphemeride
  {
  public: 
      //       
    virtual TCoordSysId GetCS() const
      { return csIdJ2000S;};

      //     
    virtual bool LoadData(const TimeJD &s, const TimeJD &e)
      { return true; };
      //  
    virtual void UnLoadData()
      {}
      //   .
    virtual void GetPosition(const TimeJD &t, Vect3 *pos = NULL, 
                             Vect3 *vel = NULL, Vect3 *acc = NULL) const
      {
      if (pos != NULL)
        *pos = 0.0;
      if (vel != NULL)
        *vel = 0.0;
      if (acc != NULL)
        *acc = 0.0;
      }
  };
///////////////////////////////////////////////////////////////////////////////
//   
///////////////////////////////////////////////////////////////////////////////
class EarthEphemeride : public IEphemeride
  {
  public: 
      //       
    virtual TCoordSysId GetCS() const
      { return csIdJ2000;};

      //     
    virtual bool LoadData(const TimeJD &s, const TimeJD &e)
      { return true; };
      //  
    virtual void UnLoadData()
      {}
      //   .
    virtual void GetPosition(const TimeJD &t, Vect3 *pos = NULL, 
                             Vect3 *vel = NULL, Vect3 *acc = NULL) const
      {
      if (pos != NULL)
        *pos = 0.0;
      if (vel != NULL)
        *vel = 0.0;
      if (acc != NULL)
        *acc = 0.0;
      }
  };
///////////////////////////////////////////////////////////////////////////////
#endif
