#include "AscNodeOrbit.h"
#include "Propagator.h"
#include "DrPeriod.h"
#include "GreenwichFrame.h"
#include "ECIFrame.h"
#include "XMLFunctions.h"
#include "mathconst.h"

static double GetZ(const Vect6 &v) {return v.z;}
static double GetVZ(const Vect6 &v) {return v.vz;}
static double GetY(const Vect6 &v) {return v.y;}
static double GetVY(const Vect6 &v) {return v.vy;}

bool AscNodeOrbit::calcAscNode(Propagator& prop, bool vzmode, PhasePoint6D& result)
{
	//at that moment propagator should be already started, not in the asc node, but in the middle of orbit

	if(vzmode){
		//if(!GetAscNodeYZ(*prop.GetPredictor()->GetPredictor(), prop.GetPredictor()->GetTime(), result, GetZ, GetVZ)){
		//	return false;
		//};

		SInitOrbit ascNodeOrbit;
		jumpToArgOfLatitude(prop, 0, ascNodeOrbit);
		result = ascNodeOrbit.GetPhasePointJ2000();

	}else{
		if(!GetAscNodeYZ(*prop.GetPredictor()->GetPredictor(), prop.GetPredictor()->GetTime(), result, GetY, GetVY)){
			return false;
		}
	}

	return true;
}

bool AscNodeOrbit::calculate(const SInitOrbit& orbit, bool vzmode)
{
	Propagator prop(orbit, false);
	
	//propagator should be started on date of orbit, so we predict orbit to 0 dt and dummy point
	PhasePoint6D dummy;
	prop.Propagate(orbit.GetDate(), dummy);
		
	PhasePoint6D j2000AscNodePoint;
	calcAscNode(prop, vzmode, j2000AscNodePoint);
	
	PhasePoint6D GCSAscNodePoint = j2000AscNodePoint;
	GCSAscNodePoint.Convert(GreenwichFrame());

	PhasePoint6D ECIAscNodePoint = j2000AscNodePoint;
	ECIAscNodePoint.Convert(ECIFrame(j2000AscNodePoint.T));

	//at that moment propagator is at ascending node
	getOrbPeriod(prop, vzmode, m_drperiod, m_drperiodDt);

	GeoCoordsPrec gcc;
	gcc.SetXYZ(GCSAscNodePoint.CoordsVel);

	m_latitude = gcc.Phi;
	m_longitude = gcc.Lam;
	m_altitude = gcc.H;

	m_kdj2000 = getKeplerData(j2000AscNodePoint.CoordsVel);
	m_kdeci = getKeplerData(ECIAscNodePoint.CoordsVel);
	m_j2000 = j2000AscNodePoint.CoordsVel;
	m_date = j2000AscNodePoint.T;
	
	return true;
}

double AscNodeOrbit::getKeplerianTime(double TA,double e,double w) const
{
	double tgE2 = tan(TA/2.0)*sqrt((1-e)/(1+e));
	double E = atan(tgE2)*2.0;
	return (E-e*sin(E))/w;
}

bool AscNodeOrbit::jumpToArgOfLatitude(Propagator& prop, double dstU, SInitOrbit& result) const
{
	PhasePoint6D p;
	int maxInter = 20;
	int iter = 0;
	bool nearestAscendingNode = false;

	if (!nearestAscendingNode){
		prop.GetPredictor()->GetPredictor()->GetDestPoint(p);
		Kepler kep;
		kep.SetXYZ(p.CoordsVel);
		TimeJD tan = kep.GetTimeFromAN(p.T);
		prop.Propagate(tan,p);
	}

	while (1){

		prop.GetPredictor()->GetPredictor()->GetDestPoint(p);
		TimeJD sjd = p.T;

		//		std::cout << DateToStr(sjd) << "\n";

		Kepler kep;
		kep.SetXYZ(p.CoordsVel);

		double residual = Distance(0,kep.U,0,dstU);
		if(fabs(residual)*rad2deg < 0.001){
			break;
		}

		if(iter++ > maxInter){
			return false;
		}

		double e = kep.GetEcc();
		double u = kep.U;
		double period = kep.GetPeriod();
		double w = 86400/period/1000*Pi2/1440/60;
		double t = getKeplerianTime((u-kep.GetArgP()),e,w);
		double t2 = getKeplerianTime((dstU-kep.GetArgP()),e,w);
		double step = t2-t;

		if(fabs(step) < 0.001){
			break;
		}

		sjd = ShiftDate(sjd, step);

		prop.Propagate(sjd, p);
	}

	if(!prop.Propagate(p.T, result)){
		return false;
	}

	return true;
}

//      
//   . 
// .   ./
bool AscNodeOrbit::getOrbPeriod(Propagator& prop, bool vzmode, double &drperiod, double &drperiodDt)
{
	//at that moment propagator should be at ascending node already

	PhasePoint6D firstAscNode;
	prop.GetPredictor()->GetPredictor()->GetDestPoint(firstAscNode);
	
	const TimeJD firstAscDate = firstAscNode.T;
	
	J2000Frame f;
	if (firstAscNode.GetFrame()->GetType() != csIdJ2000){
		firstAscNode.Convert(f);
	}

	Kepler kp;
	kp.SetXYZ(firstAscNode.CoordsVel);
	double period = kp.GetPeriod();
		
	PhasePoint6D secondAscNode;
	prop.Propagate(ShiftDate(firstAscDate, - .5 * period * 1000), secondAscNode);
	if (!calcAscNode(prop, vzmode, secondAscNode)){
		return false;
	}

	TimeJD secondAscDate = secondAscNode.T;
		
	PhasePoint6D thirdAscNode;
	prop.Propagate(ShiftDate(secondAscDate, - .5 * period * 1000), thirdAscNode);
	if (!calcAscNode(prop, vzmode, thirdAscNode)){
		return false;
	}

	TimeJD thirdAscDate = thirdAscNode.T;

	double drperiod1 = (firstAscDate - secondAscDate).GetDouble()*86400/1000;
	double drperiod2 = (secondAscDate - thirdAscDate).GetDouble()*86400/1000;
	
	drperiod = drperiod1; 
	drperiodDt = drperiod1 - drperiod2;

	return true;
};

static void PutValue(std::ostream& ss, std::string conName, const Vect6& value)
{
	ss << "<" << conName << ">";
	
	PutValue(ss, "x", value.x);
	PutValue(ss, "y", value.y);
	PutValue(ss, "z", value.z);
	PutValue(ss, "vx", value.vx);
	PutValue(ss, "vy", value.vy);
	PutValue(ss, "vz", value.vz);
	
	ss << "</" << conName << ">";
}

void AscNodeOrbit::ToXml(std::ostream& outs) const
{
	PutValue(outs,"date", m_date);
	PutValue(outs,"latitude", m_latitude);
	PutValue(outs,"longitude", m_longitude);
	PutValue(outs,"altitude", m_altitude);
	PutValue(outs,"drperiod", m_drperiod);
	PutValue(outs,"drperiodDt", m_drperiodDt);
	PutValue(outs,"kepJ2000", m_kdj2000);
	PutValue(outs,"kepEci", m_kdeci);
	PutValue(outs,"vecJ2000", m_j2000);
}

KepData AscNodeOrbit::getKeplerData(const Vect6& pos)
{
	Kepler kp;
	kp.SetXYZ(pos);

	KepData kepData;
	kepData.Init(kp);

	return kepData;
}

