﻿#include "interpolate.h"
#include "Propagator.h"
#include "GreenwichFrame.h"

template<bool dir, typename vartype> struct cmppos : public std::binary_function<vartype,vartype,bool> 
{
	bool operator()(vartype s1, vartype s2) const
	{
		const TimeJD& v1 = s1.get_date();
		const TimeJD& v2 = s2.get_date();

		return (v1 == v2) ? false : (v1 < v2)^dir;
	}
};

static int get_step(double period, double ecc){
	if(ecc < 0.1){
		return std::max((int)(period/100.0*60), 1);
	}else if(ecc < 0.2){
		return std::max((int)(period/200.0*60), 1);
	}else if(ecc < 0.4){
		return std::max((int)(period/400.0*60), 1);
	}else if(ecc < 0.6){
		return std::max((int)(period/600.0*60), 1);
	}else{
		return std::max((int)(period/1000.0*60), 1);
	}
}

static int get_step(const SInitOrbit& orbit)
{
	double period = orbit.GetKeplerData().GetPeriod()*1000/60;
	double ecc = orbit.GetKeplerData().GetEccentricity();
	return get_step(period, ecc);
}

XInterpolator::XInterpolator()
:m_ephi(0)
{

}

XInterpolator::XInterpolator(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end)
:
m_ephi(0),
m_orbit(orbit)
{
	create(orbit, beg, end);
}

XInterpolator::~XInterpolator()
{
	if(m_ephi){
		delete m_ephi;
		m_ephi = 0;
	}
}

void XInterpolator::propagate_dir(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, int dir, std::vector<IRes>& result)
{
	Propagator prop(orbit, false);
	int step = get_step(orbit);

	//std::string s_beg = DateToStr(beg);
	//std::string s_end = DateToStr(end);

	bool exit = false;
	TimeJD date = beg;
	while(!exit){
		//std::string s_date = DateToStr(date);

		if(date >= end && dir == 1 || date <= end && dir == -1){
			date = end;
			exit = true;
		}
	
		PhasePoint6D pos;
		prop.Propagate(date, pos);
		result.push_back(IRes(pos.CoordsVel, date, step));
		date = ShiftDate(date, dir * step);

		//double age = DaysInterval(prop.GetInitOrbit().GetDate(), date);
		//std::cout << age << "\n";
		//int c=8;
	}
}

void XInterpolator::propagate(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, int dir, std::vector<IRes>& result)
{
	if(dir != 0){
		propagate_dir(orbit, beg, end, dir, result);
	}else{
		propagate_dir(orbit, ShiftDate(orbit.GetDate(), +get_step(orbit) / 2), end, +1, result);
		propagate_dir(orbit, ShiftDate(orbit.GetDate(), -get_step(orbit) / 2), beg, -1, result);
	}

	std::sort(result.begin(), result.end(), cmppos<false, IRes>());
}

void XInterpolator::calc_positions(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, std::vector<IRes>& result)
{
	int n = 100;
	int step = get_step(orbit);
	TimeJD beg_ext = ShiftDate(beg, -n*step);
	TimeJD end_ext = ShiftDate(end, +n*step);

	if (orbit.GetDate() <= beg_ext){
		propagate(orbit, beg_ext, end_ext, 1, result);
	}else if (orbit.GetDate() >= end_ext){
		propagate(orbit, end_ext, beg_ext, -1, result);
	}else{
		propagate(orbit, beg_ext, end_ext, 0, result);
	}
}

static void init(ei& ephi, int size){
	int j;
//	int idir;
//	int mjd;
//	double sod;
//	double dgast = 0.e0;
	double pi;
//	char ephstr[256];

	/* initialize all the arrays, in case there is nothing to read */
	
	ephi.jdi = new double[size]();
	memset(ephi.jdi, 0, size*sizeof(double));

	ephi.jdf = new double[size]();
	memset(ephi.jdf, 0, size*sizeof(double));

	ephi.jds = new double[size]();
	memset(ephi.jds, 0, size*sizeof(double));

//	ephi.offsetjdi = new double[size]();
//	memset(ephi.offsetjdi, 0, size*sizeof(double));

//	ephi.offsetjdf = new double[size]();
//	memset(ephi.offsetjdf, 0, size*sizeof(double));
	
	for (j = 0; j < 6; j++)	{
		ephi.outv[j] = new double[size]();
		memset(ephi.outv[j], 0, size*sizeof(double));

		//ephi.backv[j] = new double[size]();
		//memset(ephi.backv[j], 0, size*sizeof(double));
	}

	for (j = 0; j < 3; j++){
		//ephi.aberoutv[j] = new double[size]();
		//memset(ephi.aberoutv[j], 0, size*sizeof(double));

		//ephi.aberbackv[j] = new double[size]();
		//memset(ephi.aberbackv[j], 0, size*sizeof(double));

		//ephi.offsetoutv[j] = new double[size]();
		//memset(ephi.offsetoutv[j], 0, size*sizeof(double));

		//ephi.offsetbackv[j] = new double[size]();
		//memset(ephi.offsetbackv[j], 0, size*sizeof(double));

		//ephi.rotangv[j] = new double[size]();
		//memset(ephi.rotangv[j], 0, size*sizeof(double));
	}

	ephi.c12 = new double[size]();
	memset(ephi.c12, 0, size*sizeof(double));

	ephi.c23 = new double[size]();
	memset(ephi.c23, 0, size*sizeof(double));

	ephi.leapflag = new int[size]();
	memset(ephi.leapflag, 0, size*sizeof(int));

//	ephi.oscrelcorr = new double[size]();
//	memset(ephi.oscrelcorr, 0, size*sizeof(double));

//	ephi.gast = new double[size]();
//	memset(ephi.gast, 0, size*sizeof(double));

//	ephi.eopxy[0] = new double[size]();
//	memset(ephi.eopxy[0], 0, size*sizeof(double));

//	ephi.eopxy[1] = new double[size]();
//	memset(ephi.eopxy[1], 0, size*sizeof(double));

//	ephi.eopdutc = new double[size]();
//	memset(ephi.eopdutc, 0, size*sizeof(double));

	ephi.nv = -1;
//	ephi.nvoff = -1;
//	ephi.nvrot = -1;
//	ephi.nveop = -1;

	pi = 4.e0 * atan (1.e0);
}

void fill(const Vect6& vec, double jdi, double jdf, ei& ephi){
	ephi.nv++;
	ephi.leapflag[ephi.nv] = 0;
	
	ephi.outv[0][ephi.nv] = vec.x*1E6; 
	ephi.outv[1][ephi.nv] = vec.y*1E6;
	ephi.outv[2][ephi.nv] = vec.z*1E6;
//	ephi.outv[3][ephi.nv] = vec.vx*1E3; 
//	ephi.outv[4][ephi.nv] = vec.vy*1E3;
//	ephi.outv[5][ephi.nv] = vec.vz*1E3;

	ephi.jdi[ephi.nv] = jdi;
	ephi.jdf[ephi.nv] = jdf;

	if (ephi.jdf[ephi.nv] > 1.0e0){
		ephi.jdi[ephi.nv] += 1.e0;
		ephi.jdf[ephi.nv] -= 1.e0;
	}

	/* In case there is not back vector (e.g., satellites)... */
//	ephi.backv[0][ephi.nv] = -ephi.outv[0][ephi.nv];
//	ephi.backv[1][ephi.nv] = -ephi.outv[1][ephi.nv];
//	ephi.backv[2][ephi.nv] = -ephi.outv[2][ephi.nv];
//	ephi.backv[3][ephi.nv] = -ephi.outv[3][ephi.nv];
//	ephi.backv[4][ephi.nv] = -ephi.outv[4][ephi.nv];
//	ephi.backv[5][ephi.nv] = -ephi.outv[5][ephi.nv];
}

void finalize(ei& ephi){
	/*     create a small usable time tag */
	(ephi.nv)++;

	for (int i = 0; i < ephi.nv; i++){
		ephi.jds[i] = ((ephi.jdf[i] - ephi.jdf[0]) +
			(ephi.jdi[i] - ephi.jdi[0])) * 86400.e0;
	}

//	(ephi.nvoff)++;

	/*for (int i = 0; i < ephi.nvoff; i++){
		ephi.offsetjds[i] = ((ephi.offsetjdf[i] - ephi.offsetjdf[0]) +
			(ephi.offsetjdi[i] -
			ephi.offsetjdi[0])) * 86400.e0;
	}*/

//	(ephi.nvrot)++;

	/*for (int i = 0; i < ephi.nvrot; i++){
		ephi.rotjds[i] = ((ephi.rotjdf[i] - ephi.rotjdf[0]) +
			(ephi.rotjdi[i] - ephi.rotjdi[0])) * 86400.e0;
	}*/

//	(ephi.nveop)++;

	/*for (int i = 0; i < ephi.nveop; i++){
		ephi.eopjds[i] = ((ephi.eopjdf[i] - ephi.eopjdf[0]) +
			(ephi.eopjdi[i] - ephi.eopjdi[0])) * 86400.e0;
	}*/
}

/*void Interpolator::filter_addons()
{
	std::vector<IRes> filtered;
	std::sort(m_positions.begin(), m_positions.end(), cmpfunc3<false, IRes>());

	int first_real;
	for (int it = 0; it < m_positions.size(); it++){
		if(!m_positions[it].is_addon()){
			first_real = it;
			break;
		}
		filtered.push_back(m_positions[it]);
	}

	int last_real;
	for (int it = m_positions.size() - 1; it >= 0; it--){
		if(!m_positions[it].is_addon()){
			last_real = it;
			break;
		}
		filtered.push_back(m_positions[it]);
	}

	for (int it = first_real; it <= last_real; it++){
		if(!m_positions[it].is_addon()){
			filtered.push_back(m_positions[it]);
		}
	}

	std::sort(filtered.begin(), filtered.end(), cmpfunc3<false, IRes>());
	m_positions = filtered;
}*/

void XInterpolator::create(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end)
{
	assert(m_ephi == 0);

	calc_positions(orbit, beg, end, m_positions);
	
	m_ephi = new ei;
	m_beg = m_positions.front().get_date();
	m_end = m_positions.back().get_date();
	
	init(*m_ephi, m_positions.size());

	for (std::vector<IRes>::const_iterator it = m_positions.begin(); it != m_positions.end(); it++){
		Vect6 pos = it->get_pos();
		TimeJD date = it->get_date();
		PhasePoint6D itrf(static_cast<const IFrame&>(J2000Frame()), pos, date);
		//itrf.Convert(GreenwichFrame());
		//double jdi = date.GetDays();
		//double jdf = date.GetFraction();
		fill(itrf.CoordsVel, date.GetDays(), date.GetFraction(), *m_ephi);
	}

	finalize(*m_ephi);
}

bool XInterpolator::get_pos(const TimeJD& date, Vect6& res) const
{
	assert(m_ephi != 0);
	assert(date >= m_beg);
	assert(date <= m_end);

	double outvec[6], backvec[6], aberout[3], aberback[3];
	double relcorrout, relcorrback;
	int fvel = 1;	
	int ircode;
	double jdi = date.GetDays();
	double sec = date.GetFraction()*86400;
	cinterp_pvac(*m_ephi, jdi, sec, outvec, backvec, aberout, aberback, &relcorrout, &relcorrback, fvel, &ircode);
	
	res = Vect6(
		outvec[0],
		outvec[1],
		outvec[2],
		outvec[3],
		outvec[4],
		outvec[5]
	);

	return true;
}
