﻿#include "orbBlocks.h"
#include "TimeFunctions.h"
#include "Propagator.h"
#include "xalgorithm.h"

template<bool dir, typename vartype> struct cmporb : public std::binary_function<vartype,vartype,bool> 
{
	bool operator()(vartype s1, vartype s2) const
	{
		TimeJD v1 = s1.GetDate();
		TimeJD v2 = s2.GetDate();

		return (v1 == v2) ? false : (v1 < v2)^dir;
	}
};

OrbBlocksStore::~OrbBlocksStore()
{
	for (std::vector<XInterpolator*>::iterator it = m_interpolators.begin(); it != m_interpolators.end(); it++){
		delete *it;
	}
	m_interpolators.clear();
}

void OrbBlocksStore::init(const SInitOrbits& orbits, const TimeJD& beg, const TimeJD& end)
{
	init_min_dist(filter(orbits), beg, end);
	
	for (OrbBlocks::const_iterator it = m_blocks.begin(); it != m_blocks.end(); it++){
		m_interpolators.push_back(new XInterpolator(it->get_orbit(), it->get_beg(), it->get_end()));
	}
}

SInitOrbits OrbBlocksStore::filter(SInitOrbits orbits) const
{
	std::sort(orbits.begin(), orbits.end(), cmporb<false, SInitOrbit>());

	SInitOrbits result;

	for (int i = 0; i < (int)orbits.size(); i++ ){
		const SInitOrbit& orb1 = orbits[i];

		TimeJD orb1beg = ShiftDate(orb1.GetDate(), -orb1.GetInterval()*86400);
		TimeJD orb1end = orb1.GetDate();

		//std::string orb1beg_s = DateToStr(orb1beg); 
		//std::string orb1end_s = DateToStr(orb1end);

		/*if (i == orbits.size() - 1){
		//result.push_back(orb1);
		break;
		}*/

		bool found = false;
		for (int j = i+1; j < (int)orbits.size(); j++ ){
			const SInitOrbit& orb2 = orbits[j];

			TimeJD orb2beg = ShiftDate(orb2.GetDate(),-orb2.GetInterval()*86400);
			TimeJD orb2end = orb2.GetDate();

			//std::string orb2beg_s = DateToStr(orb2beg); 
			//std::string orb2end_s = DateToStr(orb2end);

			bool eq1 = fabs(DaysInterval(orb1beg, orb2beg)*86400*1000) < 1;
			bool eq2 = fabs(DaysInterval(orb1end, orb2end)*86400*1000) < 1;

			//fuzzy logic consumption analisys
			if(orb1beg < orb2beg && orb1end < orb2end && orb2beg < orb1end){
				double left = DaysInterval(orb1beg, orb2beg)*86400;
				double right = DaysInterval(orb1end, orb2end)*86400;
				if(left > 0 && right > 0){
					double ratio = left / right;
					if (ratio < 0.1){
						//orbit2 consumes orbit1, so orbit1 no longer needed
						found = true;
						break;
					}
				}
			}

			//strict logic consumption analisys
			if(((orb1beg > orb2beg) || eq1) && ((orb1end < orb2end) || eq2)){
				//orbit2 consumes orbit1, so orbit1 no longer needed
				found = true;
				break;
			}
		}

		if(!found){
			result.push_back(orb1);
		}
	}

	return result;
}

void OrbBlocksStore::init_min_dist(const SInitOrbits& orbits, const TimeJD& beg, const TimeJD& end)
{
	TimeJD left = beg;
	bool stop = false;

	//	std::string beg_s = DateToStr(beg); 
	//	std::string end_s = DateToStr(end);

	for (int i = 0; i < (int)orbits.size() && !stop; i++ ){

		const SInitOrbit& orb1 = orbits[i];

		TimeJD orb1beg = ShiftDate(orb1.GetDate(),-orb1.GetInterval()*86400);
		TimeJD orb1end = orb1.GetDate();

		if (i == (int)orbits.size() - 1){
			m_blocks.push_back(OrbBlock(orb1, left, end));
			break;
		}

		//std::string orb1beg_s = DateToStr(orb1beg); 
		//std::string orb1end_s = DateToStr(orb1end);
		//std::string left_s = DateToStr(left);

		if(orb1end < left){
			continue;
		}

		const SInitOrbit& orb2 = orbits[i+1];

		TimeJD orb2beg = ShiftDate(orb2.GetDate(),-orb2.GetInterval()*86400);
		TimeJD orb2end = orb2.GetDate();

		//std::string orb2beg_s = DateToStr(orb2beg); 
		//std::string orb2end_s = DateToStr(orb2end);

		if(orb1end < orb2beg){
			//no intersection
			if(orb2beg < left){
				//orbit1 if out of left and orbit2 has left border inside, so orbit1 no longer needed
				throw (Exp() << "assert point 0");
			}else{
				TimeJD right;
				double minDist;
				calc_min_dist(orb1, orb2, orb1end, orb2beg, right, minDist);
				//std::string minjd_s = DateToStr(right);
				if(right > end){
					right = end;
					stop = true;
				}
				m_blocks.push_back(OrbBlock(orb1, left, right, minDist));
				left = right;
				continue;
			}
		}else{
			//intersection
			bool eq = fabs(DaysInterval(orb1beg, orb2beg)*86400*1000) < 1;
			if((orb1beg > orb2beg) || eq){
				throw (Exp() << "assert point 1");
			}else if(orb1beg < orb2beg){
				TimeJD right = orb2beg;
				m_blocks.push_back(OrbBlock(orb1, left, right));
				left = right;
				continue;
			}
			throw (Exp() << "assert point 2");
		}
	}
}

class MinDistEvaluator
{
public:
	MinDistEvaluator(const XInterpolator& ip1, const XInterpolator& ip2, const TimeJD& base)
		:m_ip1(ip1), m_ip2(ip2), m_base(base)
	{
	}
	double calc(double time) const {
		TimeJD date = ShiftDate(m_base, time*86400);

		Vect6 p1;
		m_ip1.get_pos(date, p1);

		Vect6 p2;
		m_ip2.get_pos(date, p2);

		return sqrt(
			sqr(p1.x - p2.x) + 
			sqr(p1.y - p2.y) + 
			sqr(p1.z - p2.z)) / 1000.0;
	}
private:
	TimeJD m_base;
	const XInterpolator& m_ip1;
	const XInterpolator& m_ip2;
};

void OrbBlocksStore::calc_min_dist(const SInitOrbit& orb1, const SInitOrbit& orb2, const TimeJD& beg, const TimeJD& end, TimeJD& result_jd, double& result_dist)
{
	XInterpolator ip1(orb1, beg, end);
	XInterpolator ip2(orb2, beg, end);
	MinDistEvaluator eval(ip1, ip2, beg);

	double step = (int)(std::min(orb1.GetKeplerData().GetPeriod(), orb2.GetKeplerData().GetPeriod()) * 1000 / 20);
	double time = 0;
	double time_end = DaysInterval(beg, end);
	double min = -1;
	double time_min = 0;

	bool exit = false;
	while(!exit){

		if(time >= time_end){
			time = time_end;
			exit = true;
		}
		
		double dist = eval.calc(time);
		
		if(min == -1 || min > dist){
			min = dist;
			time_min = time;
		}

		time += step / 86400.0;
	}

	double tol = 60 / 86400.0;
	double left = std::max(0.0, time_min - step / 86400.0 /2.0);
	double right = std::min(time_end, time_min + step / 86400.0 /2.0);
	double shift = golden_section_search(eval, left, right, tol);
	
	result_jd = ShiftDate(beg, shift * 86400);
	result_dist = eval.calc(shift);
}

bool OrbBlocksStore::get_pos(const TimeJD& date, Vect6& result) const
{
	for (std::vector<XInterpolator*>::const_iterator it = m_interpolators.begin(); it != m_interpolators.end(); it++){
		if(date >= (*it)->get_beg() && date <= (*it)->get_end()){
			return (*it)->get_pos(date, result);
		}
	}

	throw (Exp() << "no orbit found");
}

const SInitOrbit& OrbBlocksStore::get_orbit(const TimeJD& date) const
{
	for (std::vector<XInterpolator*>::const_iterator it = m_interpolators.begin(); it != m_interpolators.end(); it++){
		if(date >= (*it)->get_beg() && date <= (*it)->get_end()){
			return (*it)->get_orbit();
		}
	}

	throw (Exp() << "no orbit found");
}